package Pseudorandom_co_uk::Pseudosite::Constructor;

# Pseudosite copyright (c) 2002 Simon McVittie
# This module is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2,
# as published by the Free Software Foundation.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# Last change and base version. Please edit the last change when you change things.
# Please do not edit the base version unless you are the author.

# Base version (for patch purposes): v0.1 20020403 smcv@pseudorandom.co.uk
# Changelog since release (newest first):
# 20020403 smcv@pseudorandom.co.uk
#	Preparation for initial release.

use strict;
use warnings;
use subs qw();
use vars qw();
use utf8;

require Pseudorandom_co_uk::Pseudosite::Processor;

=head1 NAME

Pseudorandom_co_uk::Pseudosite::Constructor - base class for constructors

=head1 USAGE IN PSEUDOSITE

A constructor implements the code required to process a content "node" on a Pseudosite
website. Examples of constructors include:

NullConstructor - does absolutely nothing
CopyConstructor

=head1 PROGRAMMING INTERFACE

=head2 "Public" methods

=over

=item $constructor = new Pseudorandom_co_uk::Pseudosite::Constructor ($processor);
=item $constructor = Pseudorandom_co_uk::Pseudosite::Constructor->new ($processor);

Creates a new Constructor, attached to the Processor object $processor.
This is fairly useless; subclasses of Constructor are far more useful :-)

=cut

sub new
{
	my $class = shift;
	$class = ref $class if ref $class;
	die "no class!" unless $class;
	my $self = {
		processor => shift, # the processor calling us
		state => {},
		config => {},
	};
	bless $self, $class;
	$self->do_config;
	return $self;
}

=item $constructor->construct( $node )

"Constructs" the Pseudosite node represented by $node, by checking its dependencies,
and (if necessary) transforming the node's input file and producing an output file.

Subclasses MUST override this method.

=cut

sub construct
{
	my $self = shift;
	my $node = shift;
	die $self->my_name . ": Abstract construct() called!";
}

=item $constructor->input_extension()
=item $constructor->output_extension()

Returns the extension to be used on input or output files which don't already have one,
without the leading dot (e.g. 'html', 'jpg', 'tar.bz2').

=item $constructor->input_extension($extension_without_leading_dot)
=item $constructor->output_extension($extension_without_leading_dot)

Sets the extension to be used on input or output files which don't already have one,
returning the new value.

=cut

sub output_extension
{
	my $self = shift;
	my $val = shift;
	return undef unless ref($self);
	$self->{config}->{output_extension} = $val if defined $val;
	return $self->{config}->{output_extension};
}

sub input_extension
{
	my $self = shift;
	my $val = shift;
	return undef unless ref($self);
	$self->{config}->{input_extension} = $val if defined $val;
	return $self->{config}->{input_extension};
}

=item $constructor->index_page()

Returns the "index page" name to be applied to this constructor's output. If this is
undef or the null string, the node ID "foo/bar" will map to an output name and URL of
"foo/bar.html" (if the output_extension is set to 'html'). If index_page is set to,
say, index.html, the ID "foo/bar" will map to the output filename "foo/bar/index.html"
and the URL "foo/bar/". This is intended for use with Apache's DirectoryIndex directive
(or similar facilities in other web servers).

=item $constructor->index_page($filename)

Sets the "index page", returning the new value.

=cut

sub index_page
{
	my $self = shift;
	my $val = shift;
	return undef unless ref($self);
	$self->{config}->{index_page} = $val if defined $val;
	return $self->{config}->{index_page};
}

=item $constructor->verify($node)

Verifies various attributes of a node. If the constructor makes assumptions about the
presence of particular attributes in its construct() function, it must set them up here.
It is guaranteed to have had a chance to verify() any node passed to construct().

The default implementation is as follows:

=over

=item * If 'input' is undefined, set it the same as 'id'

=item * If 'output' is undefined, set it the same as 'id'

=item * Add a dot and the value of input_extension to 'input' unless it contains a dot
after the last forward-slash

=item * If 'output' does not contain a dot after the last forward-slash:

=over

=item - if index_page is set, set 'preferred-name' to 'output' plus a forward slash,
then set 'output' to 'output' plus a slash plus index_page

=item - otherwise, set both 'output' and 'preferred-name' to 'output' plus a dot and
the value of output_extension

=back

=item * set 'server-root' and 'server' to the parent node's equivalents, or failing
that, what the processor returns from server_root() and server()

=item * set 'url' to the fully qualified URL of this resource
(e.g. http://example.com/foo/bar.html) using 'preferred-name', 'server' and 'server-root'

=item * set 'url-local' to the absolute URL of this resource, as referred to from the
same server (e.g. /foo/bar.html) using 'preferred-name' and 'server-root'

=back

=cut

sub verify
{
	my $self = shift;
	my $node = shift;
	my $preferred_name;

	$node->attr('input',$node->attr('id')) unless $node->attr('input');

	$self->verify_input_name($node);

	$node->attr('infile',$self->processor->path($node->attr('from') or 'source') . '/' . $node->attr('input'));

	$node->attr('output',$node->attr('id')) unless $node->attr('output');

	$self->verify_output_name($node);

	$preferred_name = $node->attr('preferred-name');

	$node->attr('outfile', $self->processor->path($node->attr('to') or 'dest') . '/' . $node->attr('output'));

	$node->attr('server',$node->parent->attr('server')) unless $node->attr('server');
	$node->attr('server',$self->processor->server) unless $node->attr('server');

	$node->attr('server-root', $node->parent->attr('server-root'))
		unless $node->attr('server-root');
	$node->attr('server-root',$self->processor->server_root)
		unless $node->attr('server-root');

	$node->attr('url-local', $node->attr('server-root') . $preferred_name)
		unless $node->attr('url-local');

	$node->attr('url',
			$node->attr('server') . $node->attr('server-root') . $preferred_name)
		unless $node->attr('url');
}

=back

=head2 Internal methods useful to implementors

=over

=item $constructor->my_name()

Returns a human-readable name for this constructor class, used in error messages etc.

=cut

sub my_name { 'Abstract Constructor base class' }

=item $constructor->do_config()

Sets up sane defaults for the constructor. Called by new(), shouldn't be called
externally. do_config() may assume it's being called on a new Constructor and isn't
required to cooperate nicely with any configuration which is already there.

Subclasses' do_config procedures should usually (always?) start with:
	my $self = shift;
	$self->SUPER::do_config();

=cut

sub do_config
{
}

=item $constructor->dependencies( $node )

$node is a Pseudosite node. Returns things this node depends on (i.e. things whose
changes cause it to be out of date).

=cut

sub dependencies
{
	my $self = shift;
	my $node = shift;
	return [ $node->attr('infile'), ];
}

=item $constructor->newest_dependency_lmod( $node )

Returns the latest modification date/time of the newest of $node's dependencies.

=cut

sub newest_dependency_lmod
{
	my $self = shift;
	my $node = shift;
	my $dep;
	my $newest = 0;
	my $time;
	foreach $dep (@{$self->dependencies($node)})
	{
		$time = $self->lastmodified($dep);
		if ($newest < $time)
		{
			$newest = $time;
		}
	}
	$newest;
}

=item verify_input_name
=item verify_output_name

These are used internally by verify(). It may be possible to get the results you want
by overriding these instead of verify(). See the source code for details :-)

=cut

sub verify_input_name
{
	my $self = shift;
	my $node = shift;
	unless($node->attr('input') =~ m![^/]*\.[^/]*$!)
	{
		$node->attr('input',$node->attr('input') . '.' . $self->input_extension);
	}
}

sub verify_output_name
{
	my $self = shift;
	my $node = shift;
	unless($node->attr('output') =~ m![^/]*\.[^/]*$!)
	{
		if($self->index_page)
		{
			my $pref = $node->attr('output');
			$pref .= '/' unless $pref =~ s!\/$!!;
			$node->attr('preferred-name',$pref) unless $node->attr('preferred-name');
			$node->attr('output',$node->attr('output')
				. '/' . $self->index_page);
		} else {
			$node->attr('output',$node->attr('output')
				 . '.' . $self->output_extension);
			$node->attr('preferred-name',$node->attr('output'));
		}
	} else {
		$node->attr('preferred-name',$node->attr('output')) unless $node->attr('preferred-name');
	}
}

=item $constructor->lastmodified($filename)

Returns the "Last Modified" date/time of the given file according to local system
conventions (often the integer number of seconds since 1st January 1970).

=cut
sub lastmodified
{
	shift;
	return (stat(shift))[9];
}

=item $constructor->run(@program_and_arguments)

Runs the first argument, passing any additional arguments as command-line parameters.
Example: $c->run('make', '-f', $makefile_name);

=cut
sub run
{
	shift;
	system {$_[0]} @_;
}

=item $constructor->processor()

Returns the processor this constructor is attached to.

=cut
sub processor
{
	my $self = shift;
	return undef unless ref($self);
	return $self->{processor};
}

=item $constructor->verbosity()

Returns the verbosity level of the processor. 0 is quiet, 1 is normal, 2 is verbose,
high levels are for debugging.

=cut
sub verbosity
{
	my $self = shift;
	return $self->processor->verbosity(@_);
}

=item configurable_options

Returns a hash reference containing the options which you can configure through the
E<lt>constructor-config/E<gt> element. The keys are the name of the method to be
called, the values are array references containing the attributes of the
E<lt>constructor-config/E<gt> element which the method requires
(if all present, they are passed to the method in the order given; if not, that method
is not called for the particular E<lt>constructor-config/E<gt> element in question).

=cut
sub configurable_options
{
	my $self = shift;
	return { 'input_extension' => [ 'input-extension' ],
		'output_extension' => [ 'output-extension' ],
		'index_page' => [ 'index-page' ],
		};
}

1;
