package Pseudorandom_co_uk::Pseudosite::Processor;

# Pseudosite copyright (c) 2002 Simon McVittie
# This module is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2,
# as published by the Free Software Foundation.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# Last change and base version. Please edit the last change when you change things.
# Please do not edit the base version unless you are the author.

# Base version (for patch purposes): v0.1 20020403 smcv@pseudorandom.co.uk
# Changelog since release (newest first):
# 20020403 smcv@pseudorandom.co.uk
#	Preparation for initial release.

use strict;
use warnings;
use subs qw();
use vars qw(@_ $_);
use utf8;
use Pseudorandom_co_uk::Pseudosite::Constructor;
use Pseudorandom_co_uk::Pseudosite::PseudoTreeParser;
use Pseudorandom_co_uk::Pseudosite::PseudoTreeNode;
use Pseudorandom_co_uk::Pseudosite::PseudoTreeText;


sub new
{
	my $class = shift;
	die "no class!" unless $class;
	my $self = bless {
		state => {
			id => '', # the id of the current node?
			current_node => undef, # the current node itself
		},
		config => {
			root_node => undef, # the root node of the config files
			site_root_node => undef, # the root of the website (DIFFERENT!)
			thorough => 0,
			verbosity => 1,
		},
		constructors => {
		},
		paths => {
		},
		node_classes => {
		},
		root_script => shift,
		consts => { # Useful constants
			pseudosite_ns => 'http://www.pseudorandom.co.uk/xmlns/pseudosite',
			xhtml_ns => 'http://www.w3.org/1999/xhtml',
			pseudosite_metadata_ns => 'http://www.pseudorandom.co.uk/xmlns/pseudosite/metadata',
			pseudohtml_ns => 'http://www.pseudorandom.co.uk/xmlns/pseudosite/pseudohtml',
		},
		update_times => {
		},
	}, $class;
	return $self;
}

sub lastupdate
{
	my $self = shift;
	my $file = shift;
	my $t = $self->{update_times}->{$file};
	$t = 0 unless $t;
	return $t;
}

sub updated
{
	my $self = shift;
	my $file = shift;
	$self->{update_times}->{$file} = time();
}

sub save_times
{
	my $self = shift;
	my $file = shift;
	print "saving times to $file\n"  if $self->verbosity > 2;
	local $\ = "";
	local *FILE;
	local $_;
	open FILE, '>', $file or (warn "Couldn't open $file for writing: $!", return undef);
	foreach (keys %{$self->{update_times}})
	{
		print FILE $self->{update_times}->{$_}, " $_\n";
	}
	close FILE;
}

sub load_times
{
	my $self = shift;
	my $file = shift;
	print "loading times from $file\n"  if $self->verbosity > 2;
	local $/ = "\n";
	local $_;
	local *FILE;
	open FILE, '<', $file or die "Couldn't open $file for reading: $!";
	while(<FILE>)
	{
		chomp;
		if(m/(\d+) (.*)/)
		{
			$self->{update_times}->{$2} = $1;
		}
	}
	close FILE;
}

sub process_file
{
	my $proc = shift;
	$proc = new $proc unless ref($proc);
	my $file = shift;
	$proc->root_node(Pseudorandom_co_uk::Pseudosite::PseudoTreeParser::Parsefile($file));
	$proc->root_node->attr(' <node source file> ',$file);
	$proc->init_node($proc->root_node);

	$proc->load_times($proc->update_times_file) if ( defined $proc->update_times_file and -f $proc->update_times_file );
	$proc->process_node($proc->site_root_node);
	$proc->save_times($proc->update_times_file) if defined $proc->update_times_file;
}

sub config
{
	my $self = shift;
	my $flag = shift;
	my $val = shift;
	return undef unless ref($self);
	$self->{config}->{$flag} = $val if defined $val;
	return $self->{config}->{$flag};
}

sub root_node
{
	my $self = shift;
	my $val = shift;
	return undef unless ref($self);
	$self->{config}->{root_node} = $val if defined $val;
	return $self->{config}->{root_node};
}
sub thorough
{
	my $self = shift;
	my $val = shift;
	return undef unless ref($self);
	$self->{config}->{thorough} = $val if defined $val;
	return $self->{config}->{thorough};
}
sub update_times_file
{
	my $self = shift;
	my $val = shift;
	return undef unless ref($self);
	$self->{config}->{update_times_file} = $val if defined $val;
	return $self->{config}->{update_times_file};
}

sub default_class
{
	my $self = shift;
	my $val = shift;
	return undef unless ref($self);
	$self->{config}->{default_class} = $val if defined $val;
	return $self->{config}->{default_class};
}
sub site_root_node
{
	my $self = shift;
	my $val = shift;
	return undef unless ref($self);
	$self->{config}->{site_root_node} = $val if defined $val;
	return $self->{config}->{site_root_node};
}
sub server
{
	my $self = shift;
	my $val = shift;
	return undef unless ref($self);
	$self->{config}->{server} = $val if defined $val;
	return $self->{config}->{server};
}
sub server_root
{
	my $self = shift;
	my $val = shift;
	return undef unless ref($self);
	$self->{config}->{'server-root'} = $val if defined $val;
	return $self->{config}->{'server-root'};
}
sub current_node
{
	my $self = shift;
	my $val = shift;
	return undef unless ref($self);
	$self->{state}->{current_node} = $val if defined $val;
	return $self->{state}->{current_node};
}
sub verbosity
{
	# 0 = quiet, 1 = normal, 2 = verbose, 3 = silly
	my $self = shift;
	my $val = shift;
	return undef unless ref($self);
	$self->{config}->{'verbosity'} = $val if defined $val;
	return $self->{config}->{'verbosity'};
}

sub path
{
	my $self = shift;
	my $flag = shift;
	my $val = shift;
	return undef unless ref($self);
	$self->{paths}->{$flag} = $val if defined $val;
	return $self->{paths}->{$flag};
}

sub list_classnames
{
	my $self = shift;
	my @arr;
	return keys %{$self->{node_classes}};
}

sub list_classobjs
{
	my $self = shift;
	my @arr;
	return values %{$self->{node_classes}};
}

sub node_class
{
	my $self = shift;
	my $classname = shift;
	my $defs = shift;
	return undef unless ref($self);
	$self->{node_classes}->{$classname} = $defs if (ref $defs eq 'HASH');
	return $self->{node_classes}->{$classname};
}

sub init_node
{
	my $self = shift;
	my $node = shift;
	my $classdef;
	my $str;
	my $constructor;
	return if $node->isa("Pseudorandom_co_uk::Pseudosite::PseudoTreeText");
	die 'Not an element!' unless $node->isa("Pseudorandom_co_uk::Pseudosite::PseudoTreeNode");
	$self->current_node($node);
	die('Unknown namespace ' . $node->ns) if $node->ns ne $self->{consts}->{pseudosite_ns};
	if($node->type eq 'node')
	{
		$self->site_root_node($node) unless $self->site_root_node;
		die "node has no ID!" unless $node->attr('id');
		print "node ".$node->attr('id')."\n"  if $self->verbosity > 1;
		$node->attr('class',$self->default_class) unless $node->attr('class');

		$classdef = $self->node_class($node->attr('class'));

		die "Unknown class " . $node->attr('class') unless $classdef;

		$node->attr('style',$node->parent->attr('style'))
			if( !$node->attr('style')
				and $node->parent
				and UNIVERSAL::isa($node->parent,"Pseudorandom_co_uk::Pseudosite::PseudoTreeNode")
				and $node->parent->fulltype eq $node->fulltype
				and $node->parent->attr('class') eq $node->attr('class')
			);
		$node->attr('style',$classdef->{'style'}) unless $node->attr('style');

		$node->attr('params',$classdef->{'params'}) unless $node->attr('params');
		$node->attr('linked',$classdef->{'linked'}) unless $node->attr('linked');
		$node->attr('indexed',$classdef->{'indexed'}) unless $node->attr('indexed');

		$classdef->{constructor}->verify($node);

		$self->init_children($node);
	} elsif($node->type eq 'include') {
		my $filename = $self->path($node->attr('from') or 'source') . '/';
		$filename .= $node->attr('file');
		print "including $filename\n" if $self->verbosity > 0;
		my $subtree_root = Pseudorandom_co_uk::Pseudosite::PseudoTreeParser::Parsefile($filename);
		$subtree_root->attr(' <node source file> ',$filename);
		my $i;
		my $siblings = $node->parent->kids;
		for($i=0;$i<=$#$siblings;$i++)
		{
			if($siblings->[$i] == $node)
			{
				$siblings->[$i] = $subtree_root;
				$subtree_root->force_change_parent($node->parent);
				$self->init_node($subtree_root);
			}
		}
#	} elsif($node->fulltype eq "{".$self->{consts}->{pseudosite_ns}."}perl") {
#		print "running embedded perl in input XML...\n";
#
#		my $subtree_root;
#		{
#			local $_;
#			eval ($node->text);
#			$subtree_root = $_;
#		}
#		die qq(embedded perl in xml crashed with "$@", aborting) if $@;
#
#		$subtree_root->attr(' <node source file> ',$filename);
#		my $i;
#		my $siblings = $node->parent->kids;
#		for($i=0;$i<=$#$siblings;$i++)
#		{
#			if($siblings->[$i] == $node)
#			{
#				$siblings->[$i] = $subtree_root;
#				$subtree_root->force_change_parent($node->parent);
#				$self->init_node($subtree_root);
#			}
#		}
	} elsif($node->type eq 'config') {
		$self->default_class($node->attr('default-class')) if $node->attr('default-class');
		$self->update_times_file($node->attr('update-times-file')) if $node->attr('update-times-file');
		$self->server($node->attr('server')) if $node->attr('server');
		$self->server_root($node->attr('server-root')) if $node->attr('server-root');
		$self->verbosity($node->attr('verbosity')) if $node->attr('verbosity');
	} elsif($node->type eq 'constructor-config') {
		my $class = $node->attr('class');
		my $c = $self->node_class($class)->{constructor};
		my %conf = %{$c->configurable_options};
		my ($fnname,$at);
		my @ary;
		PARAMLOOP: foreach $fnname (keys %conf)
		{
			@ary = ();
			foreach $at (@{$conf{$fnname}})
			{
				if(defined $node->attr($at))
				{
					push @ary, $node->attr($at);
				} else {
					# a required parameter is missing, they
					# obviously want to leave this config item
					# as-is
					next PARAMLOOP;
				}
			}
			# if we got here, we have all the required attributes, so:
			$c->$fnname(@ary);
		}
		$c->input_extension($node->attr('input-extension')) if $node->attr('input-extension');
		$c->output_extension($node->attr('output-extension')) if $node->attr('output-extension');
		$c->index_page($node->attr('index-page')) if $node->attr('index-page');
	} elsif($node->fulltype eq "{".$self->{consts}->{pseudosite_ns}."}class") {
		my $class = $node->attr('name');
		my $c = $node->attr('constructor')->new ($self);
		my $style = $node->attr('style');
		my $params = $node->attr('params');
		my $i = $node->attr('indexed');
		$i = 'yes' unless defined $i;
		my $l = $node->attr('linked');
		$l = 'yes' unless defined $l;
		print "registered class: $class\n" if $self->verbosity > 1;
		print "  (style \"$style\", params \"$params\", ",
			$self->is_no($i) ? 'not' : $self->is_yes($i) ? '' : 'might be', " indexed, ",
			$self->is_no($l) ? 'not' : $self->is_yes($l) ? '' : 'might be', " linked, ",
			"constructor \"", $node->attr('constructor'), "\" => $c",
			")\n" if $self->verbosity > 2;
		$self->node_class($class,{ 'style' => $style, 'params' => $params,
			'constructor' => $c, 'indexed' => $i, 'linked' => $l, });
	} elsif($node->fulltype eq "{".$self->{consts}->{pseudosite_ns}."}location") {
		my $name = $node->attr('name');
		my $root = $node->attr('root');
		print "location: $name = $root\n" if $self->verbosity > 1;
		$self->path($name,$root);
	} elsif($node->fulltype eq "{".$self->{consts}->{pseudosite_ns}."}extension") {
		my $file = $node->attr('file');
		my $package = $node->attr('package');
		my $c;
		if($file)
		{
			print "extension: $file\n" if $self->verbosity > 1;
			local $_ = $self;
			require $file;
		} else {
			die "Couldn't load extension $file";
		}
	} elsif($node->fulltype eq "{".$self->{consts}->{pseudosite_ns}."}pseudosite") {
		$self->init_children($node);
	} else {
		die "Unknown element: " . $node->fulltype;
	}
	return $node;
}

sub init_children
{
	my $self = shift;
	my $node = shift;
	return if (!ref($node) or $node->isa("Pseudorandom_co_uk::Pseudosite::PseudoTreeText"));
	die "Not an element!" unless $node->isa("Pseudorandom_co_uk::Pseudosite::PseudoTreeNode");
	my $kid;
	my @kids = @{$node->kids};
	foreach $kid (@kids)
	{
		$self->init_node($kid);
	}
}

sub process_node
{
	my $self = shift;
	my $node = shift;
	my $constructor;
	return if (!ref($node) or $node->isa("Pseudorandom_co_uk::Pseudosite::PseudoTreeText"));
	die "Not an element!" unless $node->isa("Pseudorandom_co_uk::Pseudosite::PseudoTreeNode");
	my $kid;
	my @kids = @{$node->kids};
	foreach $kid (@kids)
	{
		$self->process_node($kid);
	}
	$self->current_node($node);
	if($node->fulltype eq "{".$self->{consts}->{pseudosite_ns}."}node")
	{
		$constructor = $self->node_class($node->attr('class'))->{constructor};
		$constructor->construct($node);
	} elsif($node->fulltype eq '{'.$self->{consts}->{pseudosite_ns}.'}pseudosite') {
	} else {
		die "Unknown element: " . $node->fulltype;
	}
	return $node;
}

sub find_node
{
	my $self = shift;
	my $id = shift;
	unless($id)
	{
		$id = '<undef>' unless defined $id;
		die "find_node called with invalid id '$id'";
	}
	return $self->find_node_below($id,$self->site_root_node);
}

sub find_node_below
{
	my $self = shift;
	my $id = shift;
	die "find_node_below called with invalid id" unless $id;
	my $root = shift;
	return undef unless $root;
	return $root if($root->attr('id') eq $id);
	my $node;
	my $hit;
	foreach $node (@{$root->kids})
	{
		next unless $node->isa('Pseudorandom_co_uk::Pseudosite::PseudoTreeNode');
		$hit = $self->find_node_below($id,$node);
		return $hit if $hit;
	}
	return undef;
}

sub is_yes
{
	my $self = shift;
	my $bool = shift;
	return 0 if $self->is_no($bool);
	return 1 if ($bool =~ m/^(y|t|0 but true)/i);
	return undef;
}
sub is_no
{
	my $self = shift;
	my $bool = shift;
	return 1 unless defined $bool;
	return 1 if ($bool =~ m/^(n|f)/i);
	return undef;
}

sub link_address
{
	my $self = shift;
	my $from = shift;
	my $to = shift;
	my $force = shift;

	$from = $self->current_node unless $from;
	local $_ = $to;
	$to = $self->find_node($_) unless UNIVERSAL::isa($to,'Pseudorandom_co_uk::Pseudosite::PseudoTreeNode');

	unless(UNIVERSAL::isa($to,'Pseudorandom_co_uk::Pseudosite::PseudoTreeNode'))
	{
		warn "Can't link to $_ - not found or invalid";
		return undef;
	}

	unless($force or $self->is_yes($to->attr('linked')))
	{
		warn $to->attr('id') . " can't be linked";
		return undef;
	}

	my $url = ( ( $from->attr('server') and $to->attr('server')
	            and $from->attr('server') eq $to->attr('server') ) ?
	         $to->attr('url-local')
	       : $to->attr('url') );

	unless($url)
	{
		warn "no URL available for link to " . $to->attr('id');
		return undef;
	}

	return $url;
}

1;
