# Pseudosite copyright (c) 2002 Simon McVittie
# This module is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2,
# as published by the Free Software Foundation.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# Last change and base version. Please edit the last change when you change things.
# Please do not edit the base version unless you are the author.

# Base version (for patch purposes): v0.1 20020403 smcv@pseudorandom.co.uk
# Changelog since release (newest first):
# 20020403 smcv@pseudorandom.co.uk
#	Preparation for initial release.

package Pseudorandom_co_uk::Pseudosite::PseudoHTMLConstructor;
use strict;
use warnings;
use subs qw();
use vars qw(@ISA);
use utf8;

require Pseudorandom_co_uk::Pseudosite::Constructor;
require Pseudorandom_co_uk::Pseudosite::XMLtoHTML;
require Pseudorandom_co_uk::Pseudosite::PseudoHTMLSimple;
@ISA = qw(Pseudorandom_co_uk::Pseudosite::PseudoHTMLSimple);

my %CONSTS = ( # Useful constants
	pseudosite_ns => 'http://www.pseudorandom.co.uk/xmlns/pseudosite',
	xhtml_ns => 'http://www.w3.org/1999/xhtml',
	pseudohtml_ns => 'http://www.pseudorandom.co.uk/xmlns/pseudosite/pseudohtml',
);

sub my_name { "PseudoHTML"; }

sub xmlproc
{
	my $self = shift;
	my $xml = shift;
	return $self->build_with_structure($xml);
}

sub do_config
{
	my $self = shift;
	$self->SUPER::do_config();
	$self->pseudohtml_container('content',1);
	$self->pseudohtml_container('resource',1); # FIXME implement these sometime
	$self->pseudohtml_container('title',1);
	$self->section_handlers('standard_metadata', \&standard_metadata) unless $self->section_handlers('standard_metadata');
}



sub build_content
{
	# Ignore the root node; go straight to <content> child nodes
	my $self = shift;
	my $tag = shift;
	my $retval = '';
	my $kid;
	if(ref($tag->kids) eq 'ARRAY')
	{
		my @kids = @{$tag->kids};
		while(@kids)
		{
			$kid = shift @kids;
			$retval .= $self->process_xml_node($kid) if ($kid->isa("Pseudorandom_co_uk::Pseudosite::PseudoTreeNode") and $kid->fulltype eq "{$CONSTS{pseudohtml_ns}}content")
		}
	}
	return $retval;
}

sub standard_metadata
{
	my $self = shift;
	my $root = shift;
	my $retval = '';
	my $temp = '';
	my $syn = '';
	my $kid;
	if(ref($root->kids) eq 'ARRAY')
	{
		my @kids = @{$root->kids};
		while(@kids)
		{
			$kid = shift @kids;
			next if $kid->isa("Pseudorandom_co_uk::Pseudosite::PseudoTreeText");
			if($kid->type eq 'metadata')
			{
				$temp .= $kid->attr('keywords');
				$syn .= $kid->attr('synopsis');
			}
		}
	}
	if($temp)
	{
		$temp = $self->xml_escape($temp);
		$retval .= "<meta name=\"keywords\" content=\"$temp\" />\n";
	} else {
		$retval .= "<meta name=\"ROBOTS\" content=\"NOINDEX\"/>\n";
	}
	if($syn)
	{
		$syn = $self->xml_escape($syn);
		$retval .= "<meta name=\"description\" content=\"$syn\" />\n";
	}
	return $retval;
}

sub build_headers
{
	my $self = shift;
	my $root = shift;
	my $retval = '';
	$retval .= $self->build_section($root,'standard_metadata');
	$retval .= $self->build_section($root,'navigation');
	$retval .= $self->build_section($root,'css_block');
	$retval .= $self->build_section($root,'extra_headers');
	return $retval;
}

sub build_section
{
	my $self = shift;
	my $root = shift;
	my $section = shift;

	my $handler = $self->style_section_handlers($self->processor->current_node->attr('style'),$section);
	return &{$handler}($self,$root) if $handler;
	$handler = $self->section_handlers($section);
	return &{$handler}($self,$root) if $handler;
	return '';
}

sub build_before_content { my $s = shift; return $s->build_section(shift,'before_content'); }
sub build_begin_content  { my $s = shift; return $s->build_section(shift,'begin_content'); }
sub build_end_content    { my $s = shift; return $s->build_section(shift,'end_content'); }
sub build_after_content  { my $s = shift; return $s->build_section(shift,'after_content'); }


=back

=head1 SEE ALSO

=over

=item * L<The Pseudosite processor|Pseudorandom_co_uk::Pseudosite::Processor>

=item * L<PseudoHTMLSimple|Pseudorandom_co_uk::Pseudosite::XMLtoHTML>

Indirect superclass: provides general infrastructure

=item * L<PseudoHTMLConstructor|Pseudorandom_co_uk::Pseudosite::PseudoHTMLSimple>

Superclass: a minimal implementation of PseudoHTML

=back

=cut

1;
