# Pseudosite copyright (c) 2002 Simon McVittie
# This module is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2,
# as published by the Free Software Foundation.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# Last change and base version. Please edit the last change when you change things.
# Please do not edit the base version unless you are the author.

# Base version (for patch purposes): v0.1 20020403 smcv@pseudorandom.co.uk
# Changelog since release (newest first):
# 20020403 smcv@pseudorandom.co.uk
#	Preparation for initial release.

package Pseudorandom_co_uk::Pseudosite::PseudoHTMLSimple;
use strict;
use warnings;
use subs qw();
use vars qw(@ISA);
use utf8;

require Pseudorandom_co_uk::Pseudosite::Constructor;
require Pseudorandom_co_uk::Pseudosite::XMLtoHTML;
@ISA = qw(Pseudorandom_co_uk::Pseudosite::XMLtoHTML);

my %CONSTS = ( # Useful constants
	pseudosite_ns => 'http://www.pseudorandom.co.uk/xmlns/pseudosite',
	xhtml_ns => 'http://www.w3.org/1999/xhtml',
	pseudohtml_ns => 'http://www.pseudorandom.co.uk/xmlns/pseudosite/pseudohtml',
);

sub do_config
{
	my $self = shift;
	$self->SUPER::do_config;

	$self->pseudohtml_handlers('if',\&process_if);
	$self->pseudohtml_handlers('unset',\&process_unset);
	$self->pseudohtml_handlers('set',\&process_set);
	$self->pseudohtml_handlers('unless',\&process_unless);
	$self->pseudohtml_handlers('l',\&process_l);
	$self->pseudohtml_handlers('e',\&process_e);
	$self->pseudohtml_handlers('comment',\&produce_comment);

	$self->pseudohtml_ignored('metadata',1);

	$self->ns_handler($CONSTS{pseudohtml_ns},\&process_pseudohtml) unless $self->ns_handler($CONSTS{pseudohtml_ns});
}

sub my_name { "PseudoHTML (simple)"; }

sub pseudohtml_handlers
{
	my $self = shift;
	my $key = shift;
	my $val = shift;
	return undef unless ref($self);
	$self->{config}->{pseudohtml_handlers}->{$key} = $val if defined $val;
	return $self->{config}->{pseudohtml_handlers}->{$key};
}
sub pseudohtml_ignored
{
	my $self = shift;
	my $key = shift;
	my $val = shift;
	return undef unless ref($self);
	$self->{config}->{pseudohtml_ignored}->{$key} = $val if defined $val;
	return $self->{config}->{pseudohtml_ignored}->{$key};
}
sub pseudohtml_container
{
	my $self = shift;
	my $key = shift;
	my $val = shift;
	return undef unless ref($self);
	$self->{config}->{pseudohtml_container}->{$key} = $val if defined $val;
	return $self->{config}->{pseudohtml_container}->{$key};
}
sub section_handlers
{
	my $self = shift;
	my $key = shift;
	my $val = shift;
	return undef unless ref($self);
	$self->{config}->{section_handlers}->{$key} = $val if defined $val;
	return $self->{config}->{section_handlers}->{$key};
}
sub style_section_handlers
{
	my $self = shift;
	my $style = shift;
	my $key = shift;
	my $val = shift;
	return undef unless ref($self);
	$self->{config}->{style_section_handlers}->{$style}->{$key} = $val if defined $val;
	return $self->{config}->{style_section_handlers}->{$style}->{$key};
}
sub style_pseudohtml_handlers
{
	my $self = shift;
	my $style = shift;
	my $key = shift;
	my $val = shift;
	return undef unless ref($self);
	$self->{config}->{style_pseudohtml_handlers}->{$style}->{$key} = $val if defined $val;
	return $self->{config}->{style_pseudohtml_handlers}->{$style}->{$key};
}


sub dependencies
{
	my $self = shift;
	my $node = shift;
	my @sdeps = $self->style_dependencies($node->attr('style')) ?
		@{$self->style_dependencies($node->attr('style'))} :
		();
	my @edeps = $self->everything_dependencies ?
		@{$self->everything_dependencies} :
		();
	my @arr = ( $node->attr('infile'), @edeps, @sdeps );
	while($node and UNIVERSAL::isa($node,'Pseudorandom_co_uk::Pseudosite::PseudoTreeNode'))
	{
		unshift @arr, $node->attr(' <node source file> ') if $node->attr(' <node source file> ');
		$node = $node->parent;

	}
	return \@arr;
}

sub add_everything_dependency
{
	my $self = shift;
	my $dep = shift;
	if($self->{config}->{everything_dependencies})
	{
		unshift( @{$self->{config}->{everything_dependencies}} , $dep);
	} else {
		$self->{config}->{everything_dependencies} = [ $dep ];
	}
}
sub add_style_dependency
{
	my $self = shift;
	my $style = shift;
	my $dep = shift;
	if($self->{config}->{style_dependencies}->{$style})
	{
		unshift( @{$self->{config}->{style_dependencies}->{$style}} , $dep);
	} else {
		$self->{config}->{style_dependencies}->{$style} = [ $dep ];
	}
}
sub style_dependencies
{
	my $self = shift;
	my $style = shift;
	if($self->{config}->{style_dependencies}->{$style})
	{
		return $self->{config}->{style_dependencies}->{$style};
	} else {
		return [];
	}
}
sub everything_dependencies
{
	my $self = shift;
	if($self->{config}->{everything_dependencies})
	{
		return $self->{config}->{everything_dependencies};
	} else {
		return [];
	}
}

sub xmlproc
{
	my $self = shift;
	my $xml = shift;
	return $self->build_content($xml);
}

sub process_pseudohtml
{
	my $self = shift;
	my $tag = shift;

	my $handler = $self->style_pseudohtml_handlers($self->processor->current_node->attr('style'),$tag->type);
	if($handler)
	{
		#print "handling ", $tag->type, " with style handler $handler\n";
		return &{$handler}($self,$tag)
	}
	if($self->pseudohtml_ignored($tag->type))
	{
		#print "ignoring ", $tag->type, "\n";
		return '';
	}
	if($self->pseudohtml_container($tag->type))
	{
		return $self->process_children_of($tag);
	}
	$handler = $self->pseudohtml_handlers($tag->type);
	if($handler)
	{
		#print "handling ", $tag->type, " with generic handler $handler\n";
		return &{$handler}($self,$tag)
	}
	if($self->is_html($tag->type))
	{
		#print "handling ", $tag->type, " as HTML\n";
		return $self->process_html($tag);
	}
	return $self->process_unknown_tag($tag);
}

sub process_if
{
	my $self = shift;
	my $tag = shift;
	my $required = $tag->attr('flag');
	my $flags = $self->processor->current_node->attr('params');
	if(index(" $flags "," $required ") != -1)
	{
		return $self->process_children_of($tag);
	} else {
		return '';
	}
}

sub process_unless
{
	my $self = shift;
	my $tag = shift;
	my $required = $tag->attr('flag');
	my $flags = $self->processor->current_node->attr('params');
	if(index(" $flags "," $required ") == -1)
	{
		return $self->process_children_of($tag);
	} else {
		return '';
	}
}

sub process_set
{
	my $self = shift;
	my $tag = shift;
	my $flag = $tag->attr('flag');
	my $flags = $self->processor->current_node->attr('params');
	if(index(" $flags "," $flag ") == -1)
	{
		$self->processor->current_node->attr('params',"$flags $flag")
	}
	return '';
}

sub process_unset
{
	my $self = shift;
	my $tag = shift;
	my $flag = $tag->attr('flag');
	my $flags = $self->processor->current_node->attr('params');
	if(index(" $flags "," $flag ") != -1)
	{
		$flags = " $flags ";
		$flags =~ s/ \Q$flag\E / /;
		$self->processor->current_node->attr('params',$flags);
	}
	return '';
}

sub process_e
{
	my $self = shift;
	my $tag = shift;
	if(UNIVERSAL::isa($tag->kids->[0],"Pseudorandom_co_uk::Pseudosite::PseudoTreeText"))
	{
		$self->state_newline(0);
		return '&'.$tag->kids->[0]->text().';';
	} else {
		warn "pseudoHTML: invalid entity tag!";
		return '<!--WARNING invalid entity tag!-->';
	}
}

sub produce_comment
{
	my $self = shift;
	my $tag = shift;
	my $t = $tag->text;
	$t =~ s!\-+!\-!g; # multiple minuses mess up comments
		# FIXME: should they be preserved anyway, since the main use of this tag
		# is likely to be embedding literal SSI in the page?
	return '<!--'.$t.'-->';
}


sub process_l
{
	my $self = shift;
	my $tag = shift;
	my $href = $self->interpret_uri($tag->attr('href'));
	my $title = $tag->attr('title');
	$self->state_newline(0);
	return "<a href=\""
			.$self->xml_escape($href)
			."\" title=\""
			.($title ? $self->xml_escape($title) : $self->xml_escape($href))
			."\">"
	       .$self->process_children_of($tag)
	       ."</a>";
}

=back

=head1 SEE ALSO

=over

=item * L<The Pseudosite processor|Pseudorandom_co_uk::Pseudosite::Processor>

=item * L<PseudoHTMLSimple|Pseudorandom_co_uk::Pseudosite::XMLtoHTML>

Superclass: provides general infrastructure

=item * L<PseudoHTMLConstructor|Pseudorandom_co_uk::Pseudosite::PseudoHTMLConstructor>

Subclass: a full implementation of PseudoHTML with proper page structure

=back

=cut

1;
