package Pseudorandom_co_uk::Pseudosite::XMLtoHTML;

# Pseudosite copyright (c) 2002 Simon McVittie
# This module is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2,
# as published by the Free Software Foundation.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# Last change and base version. Please edit the last change when you change things.
# Please do not edit the base version unless you are the author.

# Base version (for patch purposes): v0.1 20020403 smcv@pseudorandom.co.uk
# Changelog since release (newest first):
# 20020403 smcv@pseudorandom.co.uk
#	Preparation for initial release.

=head1 NAME

XMLtoHTML - base class for Pseudosite XML constructors

=head1 SYNOPSIS

package Foo::Bar::FooMarkupLanguage_to_HTML;
require Pseudorandom_co_uk::Pseudosite::XMLtoHTML;
@ISA = qw(Pseudorandom_co_uk::Pseudosite::XMLtoHTML);

=head1 DESCRIPTION

An abstract constructor for L<Pseudosite|Pseudorandom_co_uk::Pseudosite::Processor>.
C<XMLtoHTML> parses an XML input file and passes text, and tags in the XHTML namespace,
into the output. Subclasses may attach their own handlers to specific namespaces.

By default, no extra verbiage is attached to the output; however, subclasses will usually
use the built-in facilities for wrapping the output in standard HTML headers.

=cut

use strict;
use warnings;
use subs qw();
use vars qw(@ISA);
use utf8;

use Pseudorandom_co_uk::Pseudosite::Constructor;
use Pseudorandom_co_uk::Pseudosite::PseudoTreeNode;
@ISA = qw(Pseudorandom_co_uk::Pseudosite::Constructor);

my %CONSTS = ( # Useful constants
	pseudosite_ns => 'http://www.pseudorandom.co.uk/xmlns/pseudosite',
	xhtml_ns => 'http://www.w3.org/1999/xhtml',
	pseudohtml_ns => 'http://www.pseudorandom.co.uk/xmlns/pseudosite/pseudohtml',
);

my %html_uris = (
	'a href' => 1, 'form action' => 1, 'body background' => 1, 'q cite' => 1,
	'blockquote cite' => 1, 'del cite' => 1, 'ins cite' => 1, 'object classid' => 1,
	'object codebase' => 1, 'applet codebase' => 1, 'object data' => 1,
	'area href' => 1, 'link href' => 1, 'img longdesc' => 1, 'frame longdesc' => 1,
	'iframe longdesc' => 1, 'head profile' => 1, 'script src' => 1, 'img src' => 1,
	'input src' => 1, 'img usemap' => 1, 'input usemap' => 1, 'object usemap' => 1,
);

sub default_newline_before
{
	my $self = shift;
	my $t = shift;
	return 1 if $t =~ m/^(address|blockquote|div|h[1-6]|p|pre
	                      |dl|dt|dd|ol|ul|li
	                      |object|param
	                      |hr
	                      |del|ins
	                      |form|fieldset|button|label|legend|select|optgroup
	                      |option|textarea|input
	                      |caption|colgroup|table|tbody|td|tfoot|th|thead|tr|col
	                      |map|area|img
	                      |noscript|script
	                      )$/x;
	return 0;
}

sub default_newline_after
{
	my $self = shift;
	my $t = shift;
	return 1 if $t =~ m/^(address|blockquote|div|h[1-6]|p|pre|br
	                      |dl|dt|dd|ol|ul|li
	                      |object|param
	                      |hr
	                      |del|ins
	                      |form|fieldset|button|label|legend|select|optgroup
	                      |option|textarea|input
	                      |caption|colgroup|table|tbody|td|tfoot|th|thead|tr|col
	                      |map|area|img
	                      |noscript|script
	                      )$/x;
	return 0;
}

sub default_newlines_inside
{
	my $self = shift;
	my $t = shift;
	return 1 if $t =~ m/^(address|blockquote|div
	                      |dl|ol|ul
	                      |del|ins
	                      |form
	                      |table|tbody|tfoot|thead|tr
	                      )$/x;
	return 0;
}

=head2 Methods

=over

=cut

sub my_name { "XMLtoHTML"; }

sub construct
{
	my $self = shift;
	my $node = shift;

	my $in_name = $node->attr('infile');
	my $out_name = $node->attr('outfile');

	if ( !( -f $in_name) )
	{
		die "Input $in_name to " . $self->my_name . " doesn't seem to exist, aborting";
	}

	if( (! -f $out_name) or $self->processor->thorough or $self->newest_dependency_lmod($node) > $self->processor->lastupdate($out_name) )
	{
		print "    (examining xml: $in_name)\n" if $self->verbosity > 1;

		my $dir = $out_name;
		my @list;
		while( $dir =~ m!/[^/]+$! )
		{
			$dir =~ s!^(.*)/[^/]+$!$1!;
			unshift @list, $dir;
		}
		foreach $dir (@list)
		{
			unless (-d $dir or $dir eq '')
			{
				print "$dir doesn't exist, creating it...\n" if $self->verbosity > 0;
				mkdir $dir or die "Couldn't create $dir";
			}
		}
		$self->do_construct($in_name,$out_name);
		$self->processor->updated($out_name);
	}
}

sub do_construct
{
	local $\;
	local $/;

	my $self = shift;
	my $infile = shift;
	my $outfile = shift;

	my $xml = Pseudorandom_co_uk::Pseudosite::PseudoTreeParser::Parsefile($infile);
	$self->state_pre(0);
	$self->state_newline(1);
	my $out = $self->xmlproc($xml);
	my $old = '';

	if(-f $outfile)
	{
		open OUTF,"<", $outfile or die "Couldn't open $outfile for reading";
		binmode OUTF;
		undef $/;
		$old = <OUTF>;
		close OUTF;
	}

	if($old ne $out)
	{
		$/ = '';
		print $self->my_name . ": $outfile\n" if $self->verbosity > 0;
		open OUTF,">", $outfile or die "Couldn't open $outfile for writing";
		binmode OUTF;
		print OUTF $out;
		close OUTF;
	}
}

sub xmlproc
{
	my $self = shift;
	my $xml = shift;
	die "Called abstract XMLtoHTML constructor!";
}

sub do_config
{
	my $self = shift;
	$self->output_extension('html') unless $self->output_extension();
	$self->input_extension('xml') unless $self->input_extension();
	$self->ns_handler($CONSTS{xhtml_ns},\&process_html) unless $self->ns_handler($CONSTS{xhtml_ns});
	$self->newline_before_fn(\&default_newline_before) unless $self->newline_before_fn();
	$self->newline_after_fn(\&default_newline_after) unless $self->newline_after_fn();
	$self->newlines_inside_fn(\&default_newlines_inside) unless $self->newlines_inside_fn();
	$self->html_uris(\%html_uris) unless $self->html_uris;
}

=item $constructor->newline_before($string)
=item $constructor->newlines_inside($string)
=item $constructor->newline_after($string)

Returns a boolean indicating whether the HTML tag $string should be preceded by, contain
or be followed by an extra newline. newline_before controls the newline before the start
tag (or empty tag), newlines_inside controls the newlines after the start
tag and before the end tag, and newline_after controls the newline after the end tag
(or empty tag). Usage: if($c->newlines_inside('h1')) { print "\n" }

=cut

sub newline_after
{
	my $self = shift;
	return &{$self->newline_after_fn}($self,@_);
}
sub newline_before
{
	my $self = shift;
	return &{$self->newline_after_fn}($self,@_);
}
sub newlines_inside
{
	my $self = shift;
	return &{$self->newline_after_fn}($self,@_);
}


=item $constructor->newline_before_fn()
=item $constructor->newlines_inside_fn()
=item $constructor->newline_after_fn()

Returns the subroutine reference used as the implementation of the above functions.

=cut

=item $constructor->newline_before_fn($coderef)
=item $constructor->newlines_inside_fn($coderef)
=item $constructor->newline_after_fn($coderef)

Replaces the implementation of the above functions with the given subroutine reference,
which should have the same semantics as newline_before etc.

=cut

sub newline_after_fn
{
	my $self = shift;
	my $val = shift;
	return undef unless ref($self);
	$self->{config}->{newline_after} = $val if defined $val;
	return $self->{config}->{newline_after};
}

sub newline_before_fn
{
	my $self = shift;
	my $val = shift;
	return undef unless ref($self);
	$self->{config}->{newline_before} = $val if defined $val;
	return $self->{config}->{newline_before};
}
sub newlines_inside_fn
{
	my $self = shift;
	my $val = shift;
	return undef unless ref($self);
	$self->{config}->{newlines_inside} = $val if defined $val;
	return $self->{config}->{newlines_inside};
}

=item $constructor->html_is_uri($tag_and_attr)

The parameter is a HTML element name followed by a single space and an attribute name,
e.g. "a href" or "img src". Returns true if the attribute should be subjected to
Pseudosite URI replacement (see interpret_uri).

=item $constructor->interpret_uri($uri)

Returns the result of performing Pseudosite URI replacement on the given URI:

=over

=item * URIs starting with a vertical bar ("|", a.k.a. pipe or "or" symbol), are assumed
to be a Pseudosite node ID, and replaced with the URI for that node

=item * Other URIs containing invalid URI characters may be processed in various ways
in future Pseudosite versions

=item * URIs not matched by a specific rule above are output as-is

=back

=item $constructor->html_uris()

Returns a reference to a hash whose keys are HTML attributes in the format "a href",
"img src" and so on; the values are what would result from calling html_is_uri on the
corresponding key. Used internally by html_is_uri.

=item $constructor->html_uris($hashref)

Parameter is a reference to a hash whose keys are HTML attributes in the format "a href",
"img src" and so on, and whose values are what html_is_uri should return for that
attribute; html_is_uri will return undef for any attribute not present in the given
hash reference from now on.

=cut

sub html_is_uri
{
	my $self = shift;
	my $tag_attr = shift;
	return $self->html_uris($tag_attr);
}

sub interpret_uri
{
	my $self = shift;
	my $val = shift;
	my ($attr,$a, $from);
	if($val =~ m/^\|(.*)(#.*)?$/)
	{
		$val = $self->processor->link_address(0,$1,0);
		warn "no URL available for $1" unless $val;
		$val .= $2 if $2;
	}
	return $val;
}

sub html_uris
{
	my $self = shift;
	my $val = shift;
	return undef unless ref($self);
	$self->{config}->{html_uris} = $val if defined $val;
	return $self->{config}->{html_uris};
}

=item $constructor->state_pre()

Returns a true value if spacing is significant at the current stage in the HTML output.

=item $constructor->state_pre($boolean)

Sets the return value for subsequent state_pre calls; you should save the value
of state_pre() and set state_pre(1) just after outputting a <pre> or similar element's
start tag, and restore the saved value when that tag is closed.

=item $constructor->state_newline()

Returns a true value if the HTML output currently ends with a newline, hence further
extra newlines (e.g. around headings) should be suppressed.

=item $constructor->state_pre($boolean)

Sets the return value for subsequent state_newline calls; you should
set state_newline(1) just after outputting a newline, and set state_newline(0) just
after outputting anything else.

=cut

sub state_pre
{
	my $self = shift;
	my $val = shift;
	return undef unless ref($self);
	$self->{state}->{pre} = $val if defined $val;
	return $self->{state}->{pre};
}
sub state_newline
{
	my $self = shift;
	my $val = shift;
	return undef unless ref($self);
	$self->{state}->{newline} = $val if defined $val;
	return $self->{state}->{newline};
}

=item $constructor->ns_handler($namespace_uri)

Returns a subroutine reference suitable to be called with parameters ($constructor,
$element) on a PseudoTreeNode object representing an XML element in the given namespace.

=item $constructor->ns_handler($namespace_uri, $reference_to_handler)

Sets the subroutine reference which will be called with parameters ($constructor,
$element) when a PseudoTreeNode object representing an XML element in the given
namespace is processed.

=cut

sub ns_handler
{
	my $self = shift;
	my $key = shift;
	my $val = shift;
	return undef unless ref($self);
	$self->{config}->{ns_handler}->{$key} = $val if defined $val;
	return $self->{config}->{ns_handler}->{$key};
}

=item $constructor->process_xml_node($pseudo_tree_object)

Processes the given PseudoTreeNode or PseudoTreeText object (including its children if
they should be processed), returning the HTML to be output. Calls plain_text for
PseudoTreeText objects, the handler returned by ns_handler for PseudoTreeNode objects,
or process_unknown_ns for PseudoTreeNode objects whose namespace has no handler.

=cut

sub process_xml_node
{
	my $self = shift;
	my $tag = shift;
	die "Malformed; $tag is not a reference!" unless ref $tag;
	if($tag->isa("Pseudorandom_co_uk::Pseudosite::PseudoTreeText"))
	{
		return $self->plain_text($tag);
	} elsif ($tag->isa("Pseudorandom_co_uk::Pseudosite::PseudoTreeNode")) {
		my $handler = $self->ns_handler($tag->ns);
		if(ref $handler eq 'CODE')
		{
			return &$handler($self,$tag);
		} else {
			return $self->process_unknown_ns($tag);
		}
	} else {
		die "Unknown class in XML tree!";
	}
}

=item $constructor->plain_text($pseudo_tree_text)

Returns the text of the given PseudoTreeText object, escaping special characters and
normalizing whitespace where necessary (depending on state_pre and state_newline).

=cut

sub plain_text
{
	my $self = shift;
	my $text = shift;
	my $before;
	my $retval = $text->xml();

	# Preformatted text is exempt from the following
	return $retval if ($self->state_pre);

	# Strip leading whitespace on new lines
	$retval =~ s/^[\n\r \t]+//g if $self->state_newline;
	# Normalize whitespace
	$retval =~ s/[ \r\n\t]/ /sg;
	$retval =~ s/\ +/ /sg;
	# There's probably something on this line now
	$self->state_newline(0) unless $retval eq '';

	return $retval;
}

=item $constructor->process_html($pseudo_tree_node)

Returns the HTML equivalent of the given PseudoTreeNode object (in the XHTML namespace),
processing any child nodes recursively through process_children_of. If the given object
is not in the XHTML namespace, outputs its unqualified name anyway (so passing something
like the object representing E<lt>somenamespace:img src="|foo"/E<gt> will just produce
something like E<lt>img src="http://mysite.com/the_uri_of_foo"/E<gt>). There is no
error checking or valid-HTML checking!

=cut

sub process_html
{
	# Copiable non-empty HTML tags
	my $self = shift;
	my $tag = shift;
	my $retval = '';
	if( (not $self->state_pre)
		and (not $self->state_newline)
		and $self->newline_before($tag->type) )
	{
		$retval .= "\n";
	}
	$retval .= '<'.$tag->type;
	my ($attr,$val, $a, $from);
	foreach $attr ($tag->attr_list)
	{
		# Special handling for |, which is a reserved character in URIs, therefore
		# shouldn't appear in a context where anyone wants to use it literally
		# | at the beginning = treat the rest of this URI as a Pseudosite node's ID
		if($self->html_is_uri($tag->type . ' ' . $attr))
		{
			$val = $self->interpret_uri($tag->attr($attr));
		} else {
			$val = $tag->attr($attr);
		}
		$val = $self->xml_escape($val);
		$a = $self->xml_escape($attr);
		$retval .= " $a=\"$val\"";
	}
	if($self->is_empty_html($tag->type))
	{
		$retval .= ' />';
		if(@{$tag->kids})
		{
			warn $tag->type . " is meant to be empty, so I'm ignoring its content";
		}
	} else {
		$retval .= '>';
		$retval .= "\n"
			if( (not $self->state_pre)
				and $self->newlines_inside($tag->type) );
		my $savedpre = $self->state_pre;
		$self->state_pre(1)
			if($tag->type =~ m/^(pre|style|script)$/
			or ($tag->attr('xml:space') and $tag->attr('xml:space') =~ m/^preserve$/i));
		$self->state_pre(0)
			if($tag->attr('xml:space') and $tag->attr('xml:space') =~ m/^default$/i);
		$retval .= $self->process_children_of($tag);
		$self->state_pre($savedpre);
		if( (not $self->state_pre)
			and (not $self->state_newline)
			and $self->newlines_inside($tag->type) )
		{
			$retval .= "\n";
		}
		$retval .= '</'.$tag->type.'>';
	}
	if( (not $self->state_pre)
		and $self->newline_after($tag->type) )
	{
		$retval .= "\n";
		$self->state_newline(1);
	} else {
		$self->state_newline(0);
	}
	return $retval;
}

=item $constructor->process_unknown_ns($pseudo_tree_node)

Processes the PseudoTreeNode representing an XML element whose namespace is unknown.
The default implementation outputs a warning to STDERR and processes the children of
the offending element.

=cut

sub process_unknown_ns
{
	my $self = shift;
	my $tag = shift;
	warn "Unknown namespace ".$tag->ns;
	return $self->process_children_of($tag);
}

=item $constructor->process_unknown_tag($pseudo_tree_node)

Processes the PseudoTreeNode representing an XML element whose namespace is known, but
whose name is unknown (e.g. E<lt>html:foo/E<gt> where html: is bound to the XHTML namespace).
The default implementation outputs a warning to STDERR and processes the children of
the offending element.

=cut

sub process_unknown_tag
{
	my $self = shift;
	my $tag = shift;
	warn "Unknown tag ".$tag->fulltype;
	return $self->process_children_of($tag);
}

=item $constructor->xml_escape($string)

Returns the string with &, E<lt>, E<gt>, " and non-ASCII characters escaped, so it's
a safe XML string (as long as it's not used in a single-quoted attribute). Pseudosite
always uses double quotes to output attributes.

=cut

sub xml_escape
{
	my $self = shift;
	return Pseudorandom_co_uk::Pseudosite::PseudoTreeText::xml_escape(@_);
}

=item $constructor->ignore($whatever_you_like)

Returns the null string. Intended to be used as a tag or namespace handler.

=cut

sub ignore
{
	return '';
}

=item $constructor->process_children_of($pseudo_tree_node)

Returns the results of processing the child nodes of the given PseudoTreeNode object,
if any.

=cut

sub process_children_of
{
	# Process children of a tag in HTML mode.
	my $self = shift;
	my $tag = shift;
	my $retval = '';
	if(ref($tag->kids) eq 'ARRAY')
	{
		my @kids = @{$tag->kids};
		while(@kids)
		{
			$retval .= $self->process_xml_node(shift @kids);
		}
	}
	return $retval;
}

=item $constructor->is_empty_html($tag_name)

Returns true if the element with the given name is an empty element in XHTML 1.1.

=cut

sub is_empty_html
{
	shift;
	my $e = shift;
	return $e =~ m/^(br|hr|img|param|area|col|input)$/ ? 1 : 0;
}

=item $constructor->is_html($tag_name)

Returns true if the element with the given name is a content element (i.e. it
appears inside the body element) in XHTML 1.1.

=cut

sub is_html
{
	shift;
	my $el = shift;
	return $el =~ m/^(abbr|acronym|address|blockquote|cite|code|dfn
		|div|em|h[1-6]|kbd|p|pre|q|samp|span|strong|var|br # Text module
		|a # Hypertext module
		|dl|dt|dd|ol|ul|li # List module
		|object|param # Object module (param is empty)
		|b|big|i|small|sub|sup|tt|hr #Presentation
		|del|ins # edit module
		|bdo # bidi text module
		|button|fieldset|form|label|legend|select|optgroup|option|textarea|input
		 # forms module
		|caption|colgroup|table|tbody|td|tfoot|th|thead|tr|col # tables module
		|map|area|img # images
		|noscript|script # scripting module
		|ruby|rbc|rtc|rb|rt|rp # Ruby module
		)$/x ? 1 : 0;
}

=item $constructor->title_of_current()

Returns the XML-escaped title of the current Pseudosite node.

=cut

sub title_of_current
{
	my $self = shift;
	return $self->xml_escape($self->processor->current_node->attr('title'));
}

=item $constructor->build_headers($rootnode)

Returns any extra headers (meta, link, style, script tags) for the document whose root
node is represented by the PseudoTreeNode object $rootnode.
Called by build_with_structure. The default implementation returns a null string;
subclasses should override this.

=item $constructor->build_before_content($rootnode)

Returns document body text to be included before the actual content (e.g. navigation
bars and other stylistic elements)
for the document whose root node is represented by the PseudoTreeNode object $rootnode.
Called by build_with_structure. The default implementation returns a null string;
subclasses should override this.

=item $constructor->build_begin_content($rootnode)

Returns document body text to be included just before the actual content (e.g. start
tags for the table or logical division element actually containing the content)
for the document whose root node is represented by the PseudoTreeNode object $rootnode.
Called by build_with_structure. The default implementation returns a null string;
subclasses should override this.

=item $constructor->build_end_content($rootnode)

Returns document body text to be included just after the actual content (e.g. end
tags for the table or logical division element actually containing the content)
for the document whose root node is represented by the PseudoTreeNode object $rootnode.
Called by build_with_structure. The default implementation returns a null string;
subclasses should override this.

=item $constructor->build_after_content($rootnode)

Returns document body text to be included after the actual content (e.g. navigation
bars and other stylistic elements)
for the document whose root node is represented by the PseudoTreeNode object $rootnode.
Called by build_with_structure. The default implementation returns a null string;
subclasses should override this.

=cut
sub build_headers {return '';}
sub build_before_content {return '';}
sub build_begin_content {return '';}
sub build_end_content {return '';}
sub build_after_content {return '';}

=item $constructor->build_with_structure($rootnode)

Returns the HTML result of processing the document whose root node is represented by
the PseudoTreeNode object $rootnode, including full headers etc.

=cut

sub build_with_structure
{
	my $self = shift;
	my $root = shift;

	my $retval;
	my $temp;

	my $xmlpi = '<?xml version="1.0" encoding="UTF-8"?>'."\n";

	# FIXME allow other than xhtml 1.1
	# FIXME possibly log what gets used and automagically pick a DTD/namespace in future incarnations?
	my $header = '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.1//EN" "http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd">'
	              ."\n"
	              .'<html xmlns="http://www.w3.org/1999/xhtml" xml:lang="en-GB">'
	              ."\n<head>\n";

	$retval = $xmlpi.$header."<title>".$self->title_of_current()."</title>\n";

	$retval .= '<meta http-equiv="Content-Type" content="text/html; charset=utf-8" />'."\n";
	$retval .= '<meta name="MSSmartTagsPreventParsing" content="TRUE" />'."\n";

	$retval .= $self->build_headers($root);

    $retval .= "</head>\n<body>\n";

	$retval .= $self->build_before_content($root) . $self->build_begin_content($root)
		. $self->build_content($root)
		. $self->build_end_content($root) . $self->build_after_content($root);

    $retval .= "</body>\n</html>\n";

    return $retval;
}

=item $constructor->build_content($rootnode)

Returns the HTML content of the document whose root node is represented by
the PseudoTreeNode object $rootnode, not including headers, footers, etc. Also used by
build_with_structure.

=cut

sub build_content
{
	my $self = shift;
	my $root = shift;
	return $self->process_xml_node($root);
}

=back

=head1 SEE ALSO

=over

=item * L<The Pseudosite processor|Pseudorandom_co_uk::Pseudosite::Processor>

=item * L<PseudoHTMLSimple|Pseudorandom_co_uk::Pseudosite::PseudoHTMLSimple>

Subclass: a simple implementation of the author's "PseudoHTML" markup language, designed
for use in SSI fragments and the like

=item * L<PseudoHTMLConstructor|Pseudorandom_co_uk::Pseudosite::PseudoHTMLConstructor>

Subclass: a full implementation of PseudoHTML with proper page structure

=back

=cut

1;
